""" OLAE sensor fusion algorithm.
"""
import jax.numpy as jnp
from jax.scipy.linalg import block_diag
from jax import jit

from jaxitude.base import MiscUtil


@jit
def olae_get_CRPq(
    w: jnp.ndarray,
    v_b_set: jnp.ndarray,
    v_n_set: jnp.ndarray
) -> jnp.ndarray:
    """ OLAE sensor fusion method which converts fusion problem into a
        least-squares problem using Cayley transforms.

    Args:
        w (jnp.ndarray): 1xN array of weights for N measurements.
        v_b_set (jnp.ndarray): 1x3N, Nx3, or 3xN array of body frame headings.
        v_n_set (jnp.ndarray): 1x3N, Nx3, or 3xN array of inertial frame
            headings.

    Returns:
        jnp.ndarray: 1x3 array of CRP parameters q
    """
    n = len(w)
    # First, convert 3xN or 1x3N arrays into 3Nx1 arrays.
    v_b_vec = v_b_set.flatten().reshape((3 * n, 1))
    v_n_vec = v_n_set.flatten().reshape((3 * n, 1))

    # Calculate summation and difference column vectors
    s = v_b_vec + v_n_vec
    d = v_b_vec - v_n_vec

    # Calculate diagonal weight block matrix.
    w_block = weight_blockmatrix(w)

    # Calculate cross product operator stack matrix for vectors.
    s_mat = stacked_s_cpo(s, n)

    # Calculate inverse matrix.
    mat1 = jnp.linalg.inv(s_mat.T @ w_block @ s_mat)

    return (mat1 @ s_mat.T @ w_block @ d).flatten()


def weight_blockmatrix(w: jnp.ndarray) -> jnp.ndarray:
    """ Creates a block matrix of w_i * I_(3x3) blocks along diagonal.

    Args:
        w (jnp.ndarray): 1xN weight array.

    Returns:
        jnp.ndarray: 3Nx3N block weight matrix.
    """
    blocks = tuple(w_i * jnp.eye(3) for w_i in w)
    return block_diag(*blocks)


def stacked_s_cpo(s: jnp.ndarray, n: float) -> jnp.ndarray:
    """ Creates a stacked array of cross product operators for every three
        elements of 3Nx1 arrau s.

    Args:
        s (jnp.ndarray): 3Nx1 array s.
        n (float): Number of measurements.

    Returns:
        jnp.ndarray: 3Nx3 array of cross product operators stacked.
    """
    return jnp.vstack(
        tuple(MiscUtil.cpo(s[3 * i:3 * (i + 1)]) for i in range(n))
    )



""" Some useful attitude transformations.
"""
import jax.numpy as jnp
from jax import jit


@jit
def cayley_transform(X: jnp.ndarray) -> jnp.ndarray:
    """ Cayley transformation for matrix X.  X must be either skew-symmetric
        or it is orthogonal.

    Args:
        X (jnp.jndarray): Skew-symmetric or orthogonal matrix.

    Returns:
        jnp.ndarray: Corresponding orthogonal or or skew-symmetric matrix.
    """
    # must be a square matrix to work.
    n = X.shape[0]
    eye = jnp.identity(n)

    return jnp.matmul(eye - X, jnp.linalg.inv(eye + X))


@jit
def q_from_cayley(dcm: jnp.ndarray) -> jnp.ndarray:
    """ Converts a rotation matrix (dcm attribute) to q via Cayley's
    transformation.

    Args:
        dcm (jnp.ndarray): 3x3 rotation dcm matrix.

    Returns:
        jnp.ndarray: 1x3 q parameter matrix.
    """
    Q = q_from_cayley(dcm)
    return jnp.ndarray([-Q[1, 2], Q[0, 2], -Q[0, 1]])


# def extended_cayley_transform(X: jnp.ndarray) -> jnp.ndarray:
#     """ Cayley transformation for matrix X.  X must be either skew-symmetric
#         or it is orthogonal.

#     Args:
#         X (jnp.jndarray): Skew-symmetric or orthogonal matrix.

#     Returns:
#         jnp.ndarray: Corresponding orthogonal or or skew-symmetric matrix.
#     """"



"""
All attitudes will be represented by products of primitive rotations R,
where R is a rotation along some coordinate axis.  Here R1, will be a
rotation along the first coordinate component, R2 the second component,
and R3 the third. For example, a 3-2-1 (Z-X-Y) Euler angle rotation
sequence by angles (a, b, c) will be M(a,b,c) = R1(c)R2(b)R3(c).
"""
from typing import Tuple
# from functools import partial

import jax.numpy as jnp
from jax import jit


# This maps Euler type to angles, given a rotation matrix R.
eulerangle_map = {
    '131': lambda R: (
        jnp.arctan2(R[0, 2], R[0, 1]),
        jnp.arccos(R[0, 0]),
        jnp.arctan2(R[2, 0], -R[1, 0])
    ),
    '121': lambda R: (
        jnp.arctan2(R[0, 1], -R[0, 2]),
        jnp.arccos(R[0, 0]),
        jnp.arctan2(R[1, 0], R[2, 0])
    ),
    '212': lambda R: (
        jnp.arctan2(R[1, 0], R[1, 2]),
        jnp.arccos(R[1, 1]),
        jnp.arctan2(R[0, 1], -R[2, 1])
    ),
    '232': lambda R: (
        jnp.arctan2(R[1, 2], -R[1, 0]),
        jnp.arccos(R[1, 1]),
        jnp.arctan2(R[2, 1], R[0, 1])
    ),
    '323': lambda R: (
        jnp.arctan2(R[2, 1], R[2, 0]),
        jnp.arctan2(jnp.sqrt(1. - R[2, 2]**2), R[2, 2]),
        jnp.arctan2(R[1, 2], -R[0, 2])
    ),
    '313': lambda R: (
        jnp.arctan2(R[2, 0], -R[2, 1]),
        jnp.arccos(R[2, 2]),
        jnp.arctan2(R[0, 2], R[1, 2])
    ),
    '132': lambda R: (
        jnp.arctan2(R[2, 1], R[1, 1]),
        jnp.arcsin(-R[0, 1]),
        jnp.arctan2(R[0, 2], R[0, 0])
    ),
    '123': lambda R: (
        jnp.arctan2(-R[1, 2], R[2, 2]),
        jnp.arcsin(R[0, 2]),
        jnp.arctan2(R[0, 1], R[0, 0])
    ),
    '213': lambda R: (
        jnp.arctan2(R[0, 2], R[2, 2]),
        jnp.arcsin(-R[1, 2]),
        jnp.arctan2(R[1, 0], R[1, 1])
    ),
    '231': lambda R: (
        jnp.arctan2(-R[0, 2], R[0, 0]),
        jnp.arcsin(R[1, 0]),
        jnp.arctan2(-R[1, 2], R[1, 1])
    ),
    '321': lambda R: (
        jnp.arctan2(R[0, 1], R[0, 0]),
        jnp.arcsin(-R[0, 2]),
        jnp.arctan2(R[1, 2], R[2, 2])
    ),
    '312': lambda R: (
        jnp.arctan2(-R[0, 1], R[1, 1]),
        jnp.arcsin(R[2, 1]),
        jnp.arctan2(-R[2, 0], R[2, 2])
    )
}


class MiscUtil(object):
    """ Container class for miscellaneous caluclations.
    """
    @staticmethod
    @jit
    def antisym_dcm_vector(dcm: jnp.ndarray) -> jnp.ndarray:
        """ Returns [
                dcm[1, 2] - dcm[2, 1],
                dcm[2, 0] - dcm[0, 2],
                dcm[0, 1] - dcm[1, 0]
            ], a vector that is used in extracting PRV, CRP q, and
            MRP s.

        Args:
            dcm (jnp.ndarray): 3x3 dcm matrix.

        Returns:
            jnp.ndarray: 1x3 matrix
        """
        return jnp.array(
            [dcm[1, 2] - dcm[2, 1],
             dcm[2, 0] - dcm[0, 2],
             dcm[0, 1] - dcm[1, 0]]
        )

    @staticmethod
    @jit
    def cpo(v: jnp.ndarray) -> jnp.ndarray:
        """ Matrix representation of cross product operator of vector v.

        Args:
            v (jnp.ndarray): 1x3 matrix (or broadcastable) representation
            of 3D vector v.

        Returns:
            jnp.ndarray: 3x3 matrix of cross product operation.
        """
        v_f = v.flatten()
        return jnp.array(
            [[0., -v_f[2], v_f[1]],
             [v_f[2], 0., -v_f[0]],
             [-v_f[1], v_f[0], 0.]]
        )

    @staticmethod
    @jit
    def swapEuler_proper(angles: jnp.array) -> jnp.array:
        """ Swaps proper Euler angles (form i-j-i) as follows:
            angle1 -> angle1 % pi, angle2 -> -angle2,
            angle3 -> angle3 % pi.  Angles should be given in radians.

        Args:
            angles (jnp.array): 1x3 matrix of proper Euler angles.

        Returns:
            jnp.array: 1x3 matrix of swapped proper Euler angles.
        """
        return jnp.array(
            [
                angles[0] - jnp.sign(angles[0]) * jnp.pi,
                -angles[1],
                angles[2] - jnp.sign(angles[2]) * jnp.pi
            ]
        )


class PRVUtil(object):
    """ Container class for PRV calculations from dcm.
    """
    @staticmethod
    @jit
    def get_e(dcm: jnp.ndarray) -> jnp.ndarray:
        """ Calculates e from dcm.

        Args:
            dcm (jnp.ndarray): dcm 3x3 matrix

        Returns:
            jnp.ndarray: 1x3 array representation of e
        """
        phi = PRVUtil.get_phi(dcm)
        e_raw = MiscUtil.antisym_dcm_vector(dcm) * 0.5 / jnp.sin(phi)
        return e_raw / jnp.linalg.norm(e_raw)

    @staticmethod
    @jit
    def get_phi(dcm: jnp.ndarray) -> float:
        """ Calculates phi from dcm.

        Args:
            dcm (jnp.ndarray): dcm 3x3 matrix

        Returns:
            float: phi
        """
        return jnp.arccos(0.5 * (dcm[0, 0] + dcm[1, 1] + dcm[2, 2] - 1.))


class Primitive(object):
    """ Object with __call__ returning self.dcm. Base class for all
        rotations and includes transformation equations from dcm.
        Base dcm attribute is a zero rotation (identity matrix).
    """
    def __init__(self) -> None:
        # Set to identity for primitives.  Does change state in subclass
        # overides, but shouldn't cause issues later.
        self.dcm = jnp.identity(3)

    def __call__(self) -> jnp.ndarray:
        """ Returns DCM matrix.

        Returns:
            jnp.ndarray: self.dcm
        """
        return self.dcm

    def get_eig(self) -> Tuple:
        """ Wrapper to call JAX.numpy.linalg.eig().

        Returns:
            Tuple: array of eigenvalues and array of eigenvectors.
        """
        return jnp.linalg.eig(self.dcm)

    def get_prv(self) -> Tuple:
        """ Returns principle angle phi and principle vector e from
            rotation matrix.

        Returns:
            Tuple: phi, vec(e)
        """
        return PRVUtil.get_phi(self.dcm), PRVUtil.get_e(self.dcm)

    def get_prv2(self) -> Tuple:
        """ Returns principle angle phi and principle vector e from
            rotation matrix for to long rotation phi' = phi - 2pi.

        Returns:
            Tuple: phi, vec(e)
        """
        return PRVUtil.get_phi(self.dcm) - 2. * jnp.pi, PRVUtil.get_e(self.dcm)

    def get_eulerangles(self, ea_type: str) -> jnp.ndarray:
        """ Returns a 1x3 matrix of Euler angles from DCM.

        Args:
            ea_type (str): Euler angle type.  Needs to be of form
                '121', '321', etc for now.
        Returns:
            jnp.ndarray: 1x3 matrix of Euler angles
        """
        return jnp.asarray(eulerangle_map[ea_type](self.dcm))

    # @partial(jit, static_argnums=0)
    def _get_b_base(self) -> jnp.ndarray:
        """ Returns a matrix of quaternion parameters from dcm. Uses Shepard's
            method to avoid singularity at b0=0.  Doesn't decide shortest path.

        Returns:
            jnp.ndarray: 1x4 matrix of quaternion parameters.
        """
        tr = jnp.trace(self.dcm)
        step1 = jnp.array(
            [
                0.25 * (1. + tr),
                0.25 * (1. + 2. * self.dcm[0, 0] - tr),
                0.25 * (1. + 2. * self.dcm[1, 1] - tr),
                0.25 * (1. + 2. * self.dcm[2, 2] - tr)
            ]
        )

        # Important: to jit, you need to cast using jnp.ndarray.astype.
        max_i = jnp.argmax(step1).item()
        step2 = jnp.array(
            [
                0.25 * (self.dcm[1, 2] - self.dcm[2, 1]),
                0.25 * (self.dcm[2, 0] - self.dcm[0, 2]),
                0.25 * (self.dcm[0, 1] - self.dcm[1, 0]),
                0.25 * (self.dcm[1, 2] + self.dcm[2, 1]),
                0.25 * (self.dcm[2, 0] + self.dcm[0, 2]),
                0.25 * (self.dcm[1, 2] + self.dcm[2, 1]),
            ]
        )
        max_sq = jnp.sqrt(step1[max_i])
        choices = {
            0: jnp.array(
                [
                    max_sq,
                    step2[0] / max_sq,
                    step2[1] / max_sq,
                    step2[2] / max_sq
                ]
            ),
            1: jnp.array(
                [
                    step2[0] / max_sq,
                    max_sq,
                    step2[3] / max_sq,
                    step2[4] / max_sq
                ]
            ),
            2: jnp.array(
                [
                    step2[1] / max_sq,
                    step2[3] / max_sq,
                    max_sq,
                    step2[5] / max_sq
                ]
            ),
            3: jnp.array(
                [
                    step2[2] / max_sq,
                    step2[4] / max_sq,
                    step2[5] / max_sq,
                    max_sq
                ]
            )
        }
        return choices[max_i]

    def get_b_short(self) -> jnp.ndarray:
        """ Shepard's method to get b from DCM. Makes sure b0 is positive.

        Returns:
            jnp.ndarray: 1x4 matrix of quaternion parameters.
        """
        b = self._get_b_base()
        return b.at[0].set(jnp.abs(b[0]))

    def get_b_long(self) -> jnp.ndarray:
        """ Shepard's method to get b from DCM. Makes sure b0 is negative.

        Returns:
            jnp.ndarray: 1x4 matrix of quaternion parameters.
        """
        b = self._get_b_base()
        return b.at[0].set(-jnp.abs(b[0]))

    def get_q(self) -> jnp.ndarray:
        """ Gets CRP q parameters from DCM.

        Returns:
            jnp.ndarray: 1x3 matrix of CRP q parameters.
        """
        zeta_squared = jnp.trace(self.dcm) + 1.
        return MiscUtil.antisym_dcm_vector(self.dcm) / zeta_squared

    def get_s(self) -> jnp.ndarray:
        """ Gets MRP s parameters from DCM.

        Returns:
            jnp.ndarray: 1x3 matrix of MRP s parameters.
        """
        zeta = jnp.sqrt(jnp.trace(self.dcm) + 1.)
        return MiscUtil.antisym_dcm_vector(self.dcm) / zeta / (zeta + 2.)


class BaseR(Primitive):
    """ Fundamental coordinate axis rotation primitive subclass.

    """
    def __init__(self, angle: float) -> None:
        """
        Attributes:
            angle (float): Rotation angle in radians.
        """
        super().__init__()
        self.angle = angle

    def inv_copy(self):
        """ Return a new instance of the same rotation class with a negative
        angle.

        Returns:
            self.__class__: New instance of the same class but with negative
            angle.
        """
        return self.__class__(-self.angle)


class R1(BaseR):
    """ Fundamental passive rotation w.r.t. coordinate axis 1.

    Args:
        BaseR: Base class
    """
    def __init__(self, angle: float) -> None:
        """
        Attibutes:
            rotor (jnp.ndarray): Overwrites primitive definition
                appropriate for R1 rotation.
        """
        super().__init__(angle)
        self.dcm = jnp.array(
            [[1., 0., 0.],
             [0., jnp.cos(angle), jnp.sin(angle)],
             [0., -jnp.sin(angle), jnp.cos(angle)]]
        )


class R2(BaseR):
    """ Fundamental passive rotation w.r.t. coordinate axis 2.

    Args:
        BaseR: Base class
    """
    def __init__(self, angle: float) -> None:
        """
        Attibutes:
            rotor (jnp.ndarray): Overwrites primitive definition
                appropriate for R2 rotation.
        """
        super().__init__(angle)
        self.dcm = jnp.array(
            [[jnp.cos(angle), 0., -jnp.sin(angle)],
             [0., 1., 0.],
             [jnp.sin(angle), 0., jnp.cos(angle)]]
        )


class R3(BaseR):
    """ Fundamental passive rotation w.r.t. coordinate axis 3.

    Args:
        BaseR: Base class
    """
    def __init__(self, angle: float) -> None:
        """
        Attibutes:
            rotor (jnp.ndarray): Overwrites primitive definition
                appropriate for R3 rotation.
        """
        super().__init__(angle)
        self.dcm = jnp.array(
            [[jnp.cos(angle), jnp.sin(angle), 0.],
             [-jnp.sin(angle), jnp.cos(angle), 0.],
             [0., 0., 1.]]
        )


class DCM(Primitive):
    """ Custom DCM

    Args:
        BaseR: Base class
    """
    def __init__(self, matrix: jnp.ndarray) -> None:
        """ Builds custom DCM instance

        Args:
            matrix (jnp.ndarray): _description_
        """
        super().__init__()
        self.dcm = jnp.asarray(matrix)
        assert self.dcm.shape == (3, 3), 'Invalid matrix shape.'



""" QUEST Method implementation
"""
from jax import grad, jit
from jax.lax import while_loop
import jax.numpy as jnp

from jaxitude.determination.davenport import get_B
from jaxitude.base import MiscUtil


def K_eig_eq(x: float, K: jnp.ndarray) -> float:
    """ Eigenvalue equation for K matrix.  To be optimized with
        respect to lam.

    Args:
        x (float): eigenvalue.
        K (jnp.ndarray): K matrix from Davenport's method.

    Returns:
        float: Value of eigenvalue polynomial equation.
    """
    return jnp.linalg.det(K - x * jnp.eye(K.shape[0]))


def get_lam(
    w: jnp.ndarray,
    K: jnp.ndarray,
    e: float = 1e-15
) -> float:
    """ Use simple Newton-Raphson method to iterate for lam with initial guess
        equal to the sum of the weights w. Function should be jitable.

    Args:
        w (jnp.ndarray): N matrix with weights.
        K (jnp.ndarray): K matrix from Davenport's method
        e (float, optional): Desired precision. Defaults to 1e-15.

    Returns:
        float: Largest eigenvalue of K.
    """
    lam0 = jnp.array([jnp.sum(w), jnp.sum(w) + 1.])
    grad_K_eig_eq = grad(K_eig_eq)
    body_func = lambda x: jnp.array(
        [x[0] - K_eig_eq(x[0], K) / grad_K_eig_eq(x[0], K), x[0]]
    )
    cond_func = lambda x: jnp.abs(x[0] - x[1]) > e
    return while_loop(cond_func, body_func, lam0)[0]


@jit
def quest_get_CRPq(
    w: jnp.ndarray,
    v_b_set: jnp.ndarray,
    v_n_set: jnp.ndarray,
    e: float = 1e-15
) -> jnp.ndarray:
    """ Get CRP q parameters from sensor heading and weights.

    Args:
        w (jnp.ndarray): N matrix with weights.
        v_b_set (jnp.ndarray): Nx3 matrix of N body frame headings from each
            sensor.
        v_n_set (jnp.ndarray): Nx3 matrix of N inertial frame headings from
            each sensor.
        e (float, optional): Desired precision. Defaults to 1e-16.

    Returns:
        jnp.ndarray: 1x3 q parameter matrix.
    """
    B = get_B(w, v_b_set, v_n_set)
    S = B + B.T
    sigma = jnp.trace(B)
    Z = jnp.expand_dims(MiscUtil.antisym_dcm_vector(B), axis=-1)
    K = jnp.block([[sigma, Z.T], [Z, S - jnp.eye(3) * sigma]])
    lam = get_lam(w, K, e=e)
    return jnp.matmul(
        jnp.linalg.inv((lam + sigma) * jnp.eye(3) - S), Z
    ).flatten()


def quest_get_Quatb(
    w: jnp.ndarray,
    v_b_set: jnp.ndarray,
    v_n_set: jnp.ndarray,
    e: float = 1e-15
) -> jnp.ndarray:
    """ Get Euler parameters b from sensor heading and weights.

    Args:
        w (jnp.ndarray): N matrix with weights.
        v_b_set (jnp.ndarray): Nx3 matrix of N body frame headings from each
            sensor.
        v_n_set (jnp.ndarray): Nx3 matrix of N inertial frame headings from
            each sensor.
        e (float, optional): Desired precision. Defaults to 1e-16.

    Returns:
        jnp.ndarray: 1x4 b parameter matrix.
    """
    q = quest_get_CRPq(w, v_b_set, v_n_set, e=e)
    q2 = jnp.dot(q, q)
    return jnp.array(1., q[0], q[1], q[2]) / jnp.sqrt(1. + q2)



""" Triad attitude estimation technique
"""
import jax.numpy as jnp
from jax import jit


def get_triad_frame(s: jnp.ndarray, m: jnp.ndarray) -> jnp.ndarray:
    """ Calculates the triad frame from vectors s and m.
        t1 = s, t2 = s x m / |s x m|, t3 = t1 x t2.

    Args:
        s (jnp.ndarray): First measurement heading vector as 1x3 matrix.
        m (jnp.ndarray): Second measurement heading vector as 1x3 matrix.

    Returns:
        jnp.ndarray: 3x3 stack of triad frame basis vectors:
            [t1.T, t2.T, t3.T].T
    """
    # Normalize s
    t1 = s / jnp.linalg.norm(s)
    # Also, normalize m.
    t2_raw = jnp.cross(t1, m / jnp.linalg.norm(m))
    t2 = t2_raw / jnp.linalg.norm(t2_raw)
    t3_raw = jnp.cross(t1, t2)
    t3 = t3_raw / jnp.linalg.norm(t3_raw)
    return jnp.vstack([t1, t2, t3]).T


@jit
def get_triad_r(
    s_b: jnp.ndarray,
    m_b: jnp.ndarray,
    s_n: jnp.ndarray,
    m_n: jnp.ndarray
) -> jnp.ndarray:
    """ Calculates body to inertial frame DCM from heading vectors given in
        body (b) and inertial (n) frames.

    Args:
        s_b (jnp.ndarray): 1x3 matrix first heading measurement in b frame.
        m_b (jnp.ndarray): 1x3 matrix second heading measurement in b frame.
        s_n (jnp.ndarray): 1x3 matrix first heading measurement in n frame.
        m_n (jnp.ndarray): 1x3 matrix second heading measurement in n frame.

    Returns:
        jnp.ndarray: 3x3 array of body to inertial frame DCM.
    """
    BT = get_triad_frame(s_b, m_b)
    NT = get_triad_frame(s_n, m_n)
    return jnp.matmul(BT, NT.T)



"""davenport's q method for quaternion parameters
"""
import jax.numpy as jnp

from jaxitude.base import MiscUtil


def get_B(
    w: jnp.ndarray,
    v_b_set: jnp.ndarray,
    v_n_set: jnp.ndarray
) -> jnp.ndarray:
    """ Generate the intermediate B matrix for davenport's q method. Heading
        vectors should be unit vectors.

    Args:
        w (jnp.ndarray): 1xN matrix of sensor weights
        v_b_set (jnp.ndarray): Nx3 matrix of N body frame headings from each
            sensor.
        v_n_set (jnp.ndarray): Nx3 matrix of N inertial frame headings from
            each sensor.

    Returns:
        jnp.ndarray: B matrix
    """
    return sum(
        [w[i] * jnp.outer(v_b_set[i], v_n_set[i]) for i in range(w.shape[0])]
    )


def get_K(
    w: jnp.ndarray,
    v_b_set: jnp.ndarray,
    v_n_set: jnp.ndarray
) -> jnp.ndarray:
    """ Generate the intermediate K matrix for davenport's q method. Heading
        vectors should be unit vectors.

    Args:
        w (jnp.ndarray): 1xN matrix of sensor weights
        v_b_set (jnp.ndarray): Nx3 matrix of N body frame headings from each
            sensor.
        v_n_set (jnp.ndarray): Nx3 matrix of N inertial frame headings from
            each sensor.

    Returns:
        jnp.ndarray: K matrix
    """
    B = get_B(w, v_b_set, v_n_set)
    sigma = jnp.trace(B)
    S = B + B.T
    Z = jnp.expand_dims(MiscUtil.antisym_dcm_vector(B), axis=-1)

    return jnp.block([[sigma, Z.T], [Z, S - jnp.eye(3) * sigma]])


def get_g(
    beta: jnp.ndarray,
    w: jnp.ndarray,
    v_b_set: jnp.ndarray,
    v_n_set: jnp.ndarray
) -> float:
    """ Get g from sensor heading and weights. Heading vectors should be unit
        vectors.

    Args:
        beta (jnp.ndarray): 1x4 matrix representation of Euler parameters.
            Input for optimization. Not usually used directly, but provided for
            completeness.
        w (jnp.ndarray): 1xN matrix of sensor weights
        v_b_set (jnp.ndarray): Nx3 matrix of N body frame headings from each
            sensor.
        v_n_set (jnp.ndarray): Nx3 matrix of N inertial frame headings from
            each sensor.

    Returns:
        float: scalar function g.
    """
    K = get_K(w, v_b_set, v_n_set)
    return jnp.matmul(
        jnp.expand_dims(beta, axis=-1),
        jnp.matmul(K, jnp.expand_dims(beta, axis=-1).T)
    )



""" Simple attitude rate from body rate ODEs.
"""
import jax.numpy as jnp
from jax import jit

from jaxitude.base import MiscUtil


@jit
def evolve_CRP(omega: jnp.ndarray, q: jnp.ndarray) -> jnp.ndarray:
    """ Returns dq/dt given q(t) and omega(t).

    Args:
        omega (jnp.ndarray): Body angular rotation rates.
        q (jnp.ndarray): 1x3 matrix of q parameteters.

    Returns:
        jnp.ndarray: dq/dt at time t.
    """
    in_shape = q.shape

    m = jnp.array(
        [[1. + q[0]**2., q[0] * q[1] - q[2], q[0] * q[2] + q[1]],
         [q[0] * q[1] + q[2], 1. + q[1]**2., q[1] * q[2] - q[0]],
         [q[0] * q[2] - q[1], q[1] * q[2] + q[0], 1. + q[2]**2.]]
    ) * 0.5
    return jnp.dot(m, omega.reshape((3, 1))).reshape(in_shape)


@jit
def evolve_MRP(omega: jnp.ndarray, s: jnp.ndarray) -> jnp.ndarray:
    """Returns ds/dt given s(t) and omega(t).

    Args:
        omega (jnp.ndarray): Body angular rotation rates.
        s (jnp.ndarray): 1x3 matrix of s parameters.

    Returns:
        jnp.ndarray: ds/dt at time t.
    """
    in_shape = s.shape
    s2 = jnp.dot(s, s)
    m = jnp.array(
        [[1. - s2 + 2. * s[0]**2., 2. * (s[0] * s[1] - s[2]), 2. * (s[0] * s[2] + s[1])],
         [2. * (s[0] * s[1] + s[2]), 1. - s2 + 2. * s[1]**2., 2. * (s[1] * s[2] - s[0])],
         [2. * (s[0] * s[2] - s[1]), 2. * (s[1] * s[2] + s[0]), 1. - s2 + 2. * s[2]**2.]]
    ) * 0.25
    return jnp.dot(m, omega.reshape((3, 1))).reshape(in_shape)


@jit
def evolve_omega_from_MRP(s: jnp.ndarray, s_dot: jnp.ndarray) -> jnp.ndarray:
    """Returns omega given s(t), s_dot(t).

    Args:
        s (jnp.ndarray): 1x3 matrix of s parameters.
        s_dot (jnp.ndarray) jnp.ndarray: 1x3 matrix of ds/dt at time t.

    Returns:
        jnp.ndarray: Body angular rotation rates.
    """
    in_shape = s.shape
    s2 = jnp.dot(s, s)
    m = jnp.array(
        [[1. - s2 + 2. * s[0]**2., 2. * (s[0] * s[1] + s[2]), 2. * (s[0] * s[2] - s[1])],
         [2. * (s[0] * s[1] - s[2]), 1. - s2 + 2. * s[1]**2., 2. * (s[1] * s[2] + s[0])],
         [2. * (s[0] * s[2] + s[1]), 2. * (s[1] * s[2] - s[0]), 1. - s2 + 2. * s[2]**2.]]
    ) * 4. / (1. + s2)**2.
    return jnp.dot(m, s_dot.reshape((3, 1))).reshape(in_shape)


@jit
def evolve_MRP_shadow(omega: jnp.ndarray, s: jnp.ndarray) -> jnp.ndarray:
    """ Returns ds/dt from s(t) and omega(t) for the corresponding MRP shadow
        set.

    Args:
        omega (jnp.ndarray): Body angular rotation rates.
        s (jnp.ndarray): 1x3 matrix of s parameteters.

    Returns:
        jnp.ndarray: ds/dt for MRP shadow set at time t
    """
    s2 = jnp.dot(s, s)
    in_shape = s.shape
    s_dot = evolve_MRP(omega, s)
    return -s_dot / s2 + 0.5 * (1. + s2) / s2**2 * jnp.dot(
        jnp.outer(s, s), omega.reshape((3, 1))
    ).reshape(in_shape)


@jit
def evolve_quat(omega: jnp.ndarray, b: jnp.ndarray) -> jnp.ndarray:
    """ Returns db/dt given b(t) and omega(t).

    Args:
        omega (jnp.ndarray): Body angular rotation rates.
        b (jnp.ndarray): 1x4 matrix of b parameteters.

    Returns:
    jnp.ndarray: db/dt at time t.
    """
    # Append zero to omega rate vector
    zero = jnp.zeros((1, 1))
    omega4 = jnp.hstack([zero, omega])
    in_shape = b.shape

    row1 = jnp.array([b[0], -b[1], -b[2], -b[3]])
    row2 = jnp.array([b[1], b[0], -b[3], b[2]])
    row3 = jnp.array([b[2], b[3], b[0], -b[1]])
    row4 = jnp.array([b[3], -b[2], b[1], b[0]])

    # |b|^2=1 at all times, so enforcing orthonormality is important.
    m = jnp.vstack(
        [row1 / jnp.linalg.norm(row1),
         row2 / jnp.linalg.norm(row2),
         row3 / jnp.linalg.norm(row3),
         row4 / jnp.linalg.norm(row4)]
    ) * 0.5
    return jnp.dot(m, omega4.reshape((4, 1))).reshape(in_shape)


@jit
def euler_eqs(omega: jnp.array, I: jnp.array) -> jnp.array:
    """ Euler equations of motion for a rigid body w.r.t. its center of mass.
        Torques are ignored.

    Args:
        omega (jnp.array): rate vector as 1x3 matrix
        I (jnp.array): 3x3 matrix inertia tensor

    Returns:
        jnp.array: I*omega_dot vector as 1x3 matrix
    """
    I_omega_dot = -MiscUtil.cpo(omega) @ I @ jnp.expand_dims(omega, axis=-1)
    return I_omega_dot.flatten()



import jax.numpy as jnp
from jax import jit


@jit
def compose_quat(b_p: jnp.ndarray, b_pp: jnp.ndarray) -> jnp.ndarray:
    """ Adds Euler parameters directly via matrix multiplication:
        Q(b) = Q(b_pp)Q(b_p) for quaternion rotation matrix Q.

    Args:
        b_p (jnp.ndarray): First rotation parameters 1x4 matrix.
        b_pp (jnp.ndarray): second rotation parameters 1x4 matrix.

    Returns:
        jnp.ndarray: Addition of rotation parameters as 1x4 matrix.
    """
    matrix = jnp.array(
        [[b_pp[0], -b_pp[1], -b_pp[2], -b_pp[3]],
         [b_pp[1], b_pp[0], b_pp[3], -b_pp[2]],
         [b_pp[2], -b_pp[3], b_pp[0], b_pp[1]],
         [b_pp[3], b_pp[2], -b_pp[1], b_pp[0]]]
    )
    return jnp.matmul(matrix, b_p.reshape((4, 1))).flatten()


@jit
def compose_crp(q_p: jnp.ndarray, q_pp: jnp.ndarray) -> jnp.ndarray:
    """ Compose CRP parameters directly in the following order:
        R(q) = R(q_pp)R(q_p) for CRP rotation matrix R.

    Args:
        q_p (jnp.ndarray): First CRP parameters as 1x3 matrix.
        q_pp (jnp.ndarray): second CRP parameters as 1x3 matrix.

    Returns:
        jnp.ndarray: Composed CRP parameters as 1x3 matrix.
    """
    num = jnp.subtract(jnp.add(q_pp, q_p), jnp.cross(q_pp, q_p))
    denom = 1. - jnp.dot(q_p, q_pp)
    return num / denom


def relative_crp(q: jnp.ndarray, q_p: jnp.ndarray) -> jnp.ndarray:
    """ Compose CRP parameters directly in the following order:
        R(q_pp) = R(q_p)R(q)^-1 for CRP rotation matrix R.
    Args:
        q (jnp.ndarray): First CRP parameters as 1x3 matrix.
        q_p (jnp.ndarray): second CRP parameters as 1x3 matrix.

    Returns:
        jnp.ndarray: Relative CRP parameters as 1x3 matrix.
    """
    return compose_crp(-q_p, q)


def compose_mrp(s_p: jnp.ndarray, s_pp: jnp.ndarray) -> jnp.ndarray:
    """ Compose MRP parameters direclty in the following order:
        R(s) = R(s_pp)R(s_p) for MRP rotation matrix R.

    Args:
        s_p (jnp.ndarray): First MRP parameters as 1x3 matrix.
        s_pp (jnp.ndarray): Second MRP parameters as 1x3 matrix.

    Returns:
        jnp.ndarray: composed MRP parameters as 1x3 matrix.
    """
    dot_p = jnp.dot(s_p, s_p)
    dot_pp = jnp.dot(s_pp, s_pp)
    return ((1. - dot_pp) * s_p + (1. - dot_p) * s_pp - 2. * jnp.cross(s_pp, s_p)) /\
        (1. + dot_p * dot_pp - 2. * jnp.dot(s_p, s_pp))

# NOTE: In Progress.
# def compose_mrp2(
#     s_p: jnp.ndarray,
#     s_pp: jnp.ndarray,
#     tol=1e-2
# ) -> jnp.ndarray:
#     """ Compose MRP parameters direclty in the following order:
#         R(s) = R(s_pp)R(s_p) for MRP rotation matrix R. If the denominator
#         is less than tol value, either s_p or s_pp are transformed to
#         their shadow set (based on which is larger: |s_p| or |s_pp|).

#     Args:
#         s_p (jnp.ndarray): First MRP parameters as 1x3 matrix.
#         s_pp (jnp.ndarray): Second MRP parameters as 1x3 matrix.
#         tol (float): Tolerance for shadow switch.

#     Returns:
#         jnp.ndarray: composed MRP parameters as 1x3 matrix.
#     """
#     dot_p = jnp.dot(s_p, s_p)
#     dot_pp = jnp.dot(s_pp, s_pp)
#     return ((1. - dot_pp) * s_p + (1. - dot_p) * s_pp -
#             - 2. * jnp.cross(s_pp, s_p)) /\
#         (1. + dot_p * dot_pp - 2. * jnp.dot(s_p, s_pp))


def relative_mrp(s: jnp.ndarray, s_p: jnp.ndarray, tol=1e-2) -> jnp.ndarray:
    """ Compose MRP parameters directly in the following order:
        R(s_pp) = R(s_p)R(s)^-1 for MRP rotation matrix R.
    Args:
        s (jnp.ndarray): First MRP parameters as 1x3 matrix.
        s_p (jnp.ndarray): Second MRP parameters as 1x3 matrix.
        tol (float): Tolerance for shadow switch.


    Returns:
        jnp.ndarray: Relative MRP parameters as 1x3 matrix.
    """
    return compose_mrp(-s_p, s, tol=tol)



"""
PRV: Principle Rotation Vector.
"""
from functools import partial

import jax.numpy as jnp
from jax import jit

from jaxitude.base import Primitive


class PRV(Primitive):
    """ Principle rotation vector object.
    """
    def __init__(self, phi: float, e: jnp.ndarray) -> None:
        """
        Attributes:
            phi (float): principle rotation in radians.
            e (jax.ndarray): 3x1 array reprentation of principle axis.
            dcm (jax.ndarray): 3x3 rotation matrix.
        """
        super().__init__()
        self.phi = phi
        self.e = e
        self.dcm = self._build_prv(phi, e)

    @partial(jit, static_argnums=0)
    def _build_prv(self, phi: float, e: jnp.ndarray):
        """ Takes input phi scalar and e vector to build PRV rotation matrix.

        Args:
            phi (float): principle rotation in radians.
            e (jax.ndarray): 3x1 array reprentation of principle axis.

        Returns:
            jnp.ndarray: PRV 3x3 dcm.
        """
        cosphi = jnp.cos(phi)
        sinphi = jnp.sin(phi)
        sigma = 1. - cosphi
        e1, e2, e3 = e.copy().flatten()
        return jnp.array(
            [[e1**2. * sigma + cosphi, e1 * e2 * sigma + e3 * sinphi, e1 * e3 * sigma - e2 * sinphi],
             [e1 * e2 * sigma - e3 * sinphi, e2**2. * sigma + cosphi, e2 * e3 * sigma + e1 * sinphi],
             [e1 * e3 * sigma + e2 * sinphi, e2 * e3 * sigma - e1 * sinphi, e3**2. * sigma + cosphi]]
        )

    def get_b_from_PVR(self) -> jnp.ndarray:
        """ Calculates and returns Euler parameters directly from phi and e.

        Returns:
            jnp.ndarray: 1x3 matrix of Euler parameters b.
        """
        return jnp.ndarray(
            [jnp.cos(self.phi / 2.),
             self.e[0] * jnp.sin(self.phi / 2.),
             self.e[1] * jnp.sin(self.phi / 2.),
             self.e[2] * jnp.sin(self.phi / 2.)]
        )

    def get_q_from_PVR(self) -> jnp.ndarray:
        """ Calculates and returns CRP q directly from phi and e.

        Returns:
            jnp.ndarray: 1x3 matrix of CRP q values.
        """
        return jnp.ndarray(
            [self.e[0] * jnp.tan(self.phi / 2.),
             self.e[1] * jnp.tan(self.phi / 2.),
             self.e[2] * jnp.tan(self.phi / 2.)]
        )

    def get_s_from_PVR(self) -> jnp.ndarray:
        """ Calculates and returns MRP s directly from phi and e.

        Returns:
            jnp.ndarray: 1x3 matrix of MRP s values.
        """
        return jnp.ndarray(
            [self.e[0] * jnp.tan(self.phi / 4.),
             self.e[1] * jnp.tan(self.phi / 4.),
             self.e[2] * jnp.tan(self.phi / 4.)]
        )



from typing import Tuple
from functools import partial

import jax.numpy as jnp
from jax import jit

from jaxitude.base import Primitive


class Quaternion(Primitive):
    """ quaternion rotation object. b0 is the scalar part and must
        be the first element of the input tuple b.
    """
    def __init__(self, b: jnp.ndarray) -> None:
        """

        Args:
            b (jnp.ndarray): 1x4 matrix of quaternion parameters. First
                component is the scalar component

        Attributes:
            b (jnp.ndarray): 1x4 matrix of quaternion parameters. First
                component is the scalar component
        """
        super().__init__()
        assert jnp.abs(jnp.linalg.norm(b) - 1.) < 1e-7, \
            'Elements of b must have unit length.'
        self.b = b
        self.dcm = self._build_quaternion(b)

    @partial(jit, static_argnums=0)
    def _build_quaternion(self, b: jnp.ndarray) -> jnp.ndarray:
        """ Builds dcm from quaternion parameters.

        Args:
            b (jnp.ndarray): 1x4 matrix parameters. First component
                is the scalar component

        Returns:
            jnp.ndarray: dcm derived from quanternion parameters b.
        """
        # Need to explicitly unpack for this to be jittable.
        b0 = b[0]
        b1 = b[1]
        b2 = b[2]
        b3 = b[3]

        return jnp.array(
            [[b0**2. + b1**2. - b2**2. - b3**2., 2. * (b1 * b2 + b0 * b3), 2. * (b1 * b3 - b0 * b2)],
             [2. * (b1 * b2 - b0 * b3), b0**2. - b1**2. + b2**2. - b3**2., 2. * (b2 * b3 + b0 * b1)],
             [2. * (b1 * b3 + b0 * b2), 2. * (b2 * b3 - b0 * b1), b0**2. - b1**2. - b2**2. + b3**2.]]
        )

    def get_PVR_from_b(self) -> Tuple:
        """ Generates PVR directly from b.

        Returns:
            tuple: scalar phi and 3x1 matrix e.
        """
        theta = 2. * jnp.arccos(self.b[0])
        e = jnp.array(
            [[self.b[1] / jnp.sin(theta)],
             [self.b[2] / jnp.sin(theta)],
             [self.b[3] / jnp.sin(theta)]]
        )
        return theta, e

    def get_q_from_b(self) -> jnp.ndarray:
        """ Generates CRP q vector from b.

        Returns:
            jnp.ndarray: 1x3 matrix of CRP q parameters.
        """
        return jnp.array(
            [self.b[1] / self.b[0],
             self.b[2] / self.b[0],
             self.b[3] / self.b[0]]
        )

    def get_s_from_b(self) -> jnp.ndarray:
        """ Generate MRP s vector from b.

        Returns:
            jnp.ndarray: 1x3 matrix of MRP s parameters.
        """
        return jnp.array(
            [self.b[1] / (1. + self.b[0]),
             self.b[2] / (1. + self.b[0]),
             self.b[3] / (1. + self.b[0])]
        )



""" CRP: Classical Rodrigues Parameters: q_i = b_i / b_0.
    MRP: Modified Rodrigues Parameters: s_i = b_i / (1 + b_0).
"""
from functools import partial

import jax.numpy as jnp
from jax import jit

from jaxitude.base import Primitive


class CRP(Primitive):
    """ Classical Rodrigues parameter rotation class.
    """
    def __init__(self, q: jnp.ndarray) -> None:
        super().__init__()
        self.q = q
        self.dcm = self._build_crp(q)

    @partial(jit, static_argnums=0)
    def _build_crp(self, q: jnp.ndarray) -> jnp.ndarray:
        """ Builds dcm from CRP q parameters.

        Args:
            q (jnp.ndarray): 1x3 matrix of CRP parameters

        Returns:
            jnp.ndarray: 3x3 rotation matrix
        """
        c = 1. + jnp.dot(q, q)
        return jnp.array(
            [[1. + q[0]**2. - q[1]**2. - q[2]**2., 2. * (q[0] * q[1] + q[2]), 2. * (q[0] * q[2] - q[1])],
             [2. * (q[0] * q[1] - q[2]), 1. - q[0]**2. + q[1]**2. - q[2]**2., 2. * (q[1] * q[2] + q[0])],
             [2. * (q[0] * q[2] + q[1]), 2. * (q[1] * q[2] - q[0]), 1. - q[0]**2. - q[1]**2. + q[2]**2.]]
        ) / c

    def get_quat_from_q(self) -> jnp.ndarray:
        """ Builds Euler parameters from CRP q.

        Returns:
            jnp.ndarray: 1x4 matrix of Euler parameters b.
        """
        q_dot_q = jnp.dot(self.q, self.q)
        b0 = 1. / jnp.sqrt(1. + q_dot_q)
        return jnp.array(
            [b0,
             self.q[0] * b0,
             self.q[1] * b0,
             self.q[2] * b0]
        )

    def get_s_from_q(self) -> jnp.ndarray:
        """ Build MRP s from CRP q.

        Returns:
            jnp.ndarray: 1x3 matrix of MRP s parameters
        """
        denom = 1. + jnp.sqrt(1. + jnp.dot(self.q, self.q))
        return self.q / denom

    def inv_copy(self):
        """ Returns new instance of CRP with q -> -q.

        Returns:
            CRP: new instance inverse CRP.
        """
        return self.__class__(-self.q)


class MRP(Primitive):
    """ Classical Rodrigues parameter rotation class.
    """
    def __init__(self, s: jnp.ndarray) -> None:
        super().__init__()
        self.s = s
        self.dcm = self._build_mrp(s)

    def get_quat_from_s(self) -> jnp.ndarray:
        """ Builds Euler parameters from MRP s.

        Returns:
            jnp.ndarray: 1x4 matrix of Euler parameters b.
        """
        s_dot_s = jnp.dot(self.s, self.s)
        c = 1. / (1. + s_dot_s)
        return jnp.array(
            [(1. - s_dot_s) * c,
             2. * self.s[0] * c,
             2. * self.s[1] * c,
             2. * self.s[2] * c]
        )

    def get_q_from_s(self) -> jnp.ndarray:
        """ Build CRP q from MRP s.

        Returns:
            jnp.ndarray: 1x3 matrix of CRP q parameters:
        """
        denom = 1. - jnp.dot(self.s, self.s)
        return 2. * self.s / denom

    @partial(jit, static_argnums=0)
    def _build_mrp(self, s: jnp.ndarray) -> jnp.ndarray:
        """ Builds dcm from MRP s parameters.

        Args:
            s (jnp.ndarray): 1x3 matrix of MRP s parameters

        Returns:
            jnp.ndarray: 3x3 rotation matrix
        """
        c = 1. - jnp.dot(s, s)
        return jnp.array(
            [[4. * (s[0]**2. - s[1]**2. - s[2]**2.) + c**2., 8. * s[0] * s[1] + 4. * c * s[2], 8. * s[0] * s[2] - 4. * c * s[1]],
             [8. * s[0] * s[1] - 4. * c * s[2], 4. * (- s[0]**2. + s[1]**2. - s[2]**2.) + c**2., 8. * s[1] * s[2] + 4. * c * s[0]],
             [8. * s[0] * s[2] + 4. * c * s[1], 8. * s[1] * s[2] - 4. * c * s[0], 4. * (- s[0]**2. - s[1]**2. + s[2]**2.) + c**2.]]
        ) / (1. + jnp.dot(s, s))**2.

    def inv_copy(self):
        """ Returns new instance of MRP with q -> -q.

        Returns:
            MRP: new instance inverse MRP.
        """
        return self.__class__(-self.s)

    def get_shadow(self):
        """ Returns shadow set instance of MRP class.

        Returns:
            CRP: new instance shadow MRP.
        """
        s_shadow = MRP.shadow(self.s)
        return self.__class__(s_shadow)

    @staticmethod
    @jit
    def shadow(s: jnp.array) -> jnp.array:
        """ Calculates shadow set from 1x3 matrix of MRP s values.

        Args:
            s (jnp.array): MRP s 1x3 matrix.

        Returns:
            jnp.array: Shadow set of MRP s 1x3 matrix.
        """
        return -s / jnp.dot(s, s)



from typing import Tuple

import jax.numpy as jnp

from jaxitude.base import R1, R2, R3, Primitive, MiscUtil


class EulerAngle(Primitive):
    """ Euler angle rotation sequence class.  Rotations are intrinsic.
    """
    def __init__(self, angles: jnp.ndarray, order: str) -> None:
        """
        Attributes:
            angles (jnp.ndarray): 1x3 matrix of Euler angles alpha, beta,
                and gamma.
            order (str): Order of rotations as a string.  An example includes
                '121', 'xyx', or 'XYX for proper x-axis, y-axis, x-axis
                Euler angles.
            self.R_alpha (PrimitiveR): Rotation object for first rotation.
            self.R_beta (PrimitiveR): Rotation object for second rotation.
            self.R_gamma (PrimitiveR): Rotation object for third rotation.
            self.dcm (jnp.ndarray): Euler angle DCM.
        """
        # Define attributes.
        self.angles = angles
        self.order = self._order_decipher(order)
        f_alpha, f_beta, f_gamma = self._order_rotations(order)
        self.R_alpha = f_alpha(angles[0])
        self.R_beta = f_beta(angles[1])
        self.R_gamma = f_gamma(angles[2])
        self.dcm = self.R_gamma() @ self.R_beta() @ self.R_alpha()

    def _order_decipher(self, order: str) -> str:
        """ Function to map input string to proper string format.

        Arguments:
            order (str): Order of rotations as a string.  An example includes
                '121', 'xyx', or 'XYX for proper x-axis, y-axis, x-axis
                Euler angles.

        Returns:
            str: order but in proper format.
        """
        mapper = {
            'x': '1',
            'y': '2',
            'z': '3',
            '1': '1',
            '2': '2',
            '3': '3'
        }

        return ''.join([mapper[i] for i in order.lower()])

    def _order_rotations(self, order: str) -> Tuple:
        """ Method to map input string to rotation order.

        Arguments:
            order (_type_): Order of rotations as a string.  An example
                includes '121', 'xyx', or 'XYX for proper x-axis, y-axis,
                x-axis Euler angles.

        Returns:
            Tuple: rotations of proper type per order.
        """
        order_proper = self._order_decipher(order)
        mapper = {
            '1': R1,
            '2': R2,
            '3': R3
        }

        return (mapper[i] for i in order_proper)

    def swapEuler_proper(self):
        """ Method to swap proper Euler angles with their valid counterparts:
            angle1 -> angle1 - pi, angle2 -> -angle2,
            angle3 -> angle3 - pi.  Returns a new EulerAngle object.

        Returns:
            EulerAngle: New EulerAngle object with swapped Euler angles.
        """
        # First, make sure the angle order is a proper Euler angle set.
        try:
            assert self.order[0] == self.order[2]

        except AssertionError:
            raise TypeError(
                f'Order {self.order} is not a proper Euler angle set.'
            )

        return self.__class__(
            MiscUtil.swapEuler_proper(self.angles), self.order
        )
